// COPYRIGHT 1994 A.D. Software, All rights reserved

// OOFILE database field-related classes
// see also oof4.cpp

#include "oof2.hpp"
#include "oof3.hpp"
#include "oofquery.hpp" 
#include "oofrel.hpp"  

// -------------------------------------------------------
//                      d b F i e l d 
// -------------------------------------------------------


dbField::dbField() :
				mBackend(0),
				mTable(0),
				mFieldNumber(0),
				mIndexOptions(kNotIndexed)
{
		dbTable::sCurrentlyConstructing->attachfield(this);
}


dbField::dbField(const OOF_IndexOptions indexOptions) :
				mBackend(0),
				mTable(0),
				mFieldNumber(0),
				mIndexOptions(indexOptions)
{
		dbTable::sCurrentlyConstructing->attachfield(this);
}


dbField::dbField(const char *fieldName) :
				mBackend(0),
				mTable(0),
				mFieldNumber(0),
				mIndexOptions(kNotIndexed),
				mFieldName(fieldName)
{
	dbTable::sCurrentlyConstructing->attachfield(this);
}


dbField::dbField(const char *fieldName, const OOF_IndexOptions indexOptions) :
				mBackend(0),
				mTable(0),
				mFieldNumber(0),
				mIndexOptions(indexOptions),
				mFieldName(fieldName)
{
	dbTable::sCurrentlyConstructing->attachfield(this);
}


dbField::dbField(const dbField& rhs) :
			dbClass(rhs),
			mBackend(rhs.mBackend),
			mTable(0),
			mFieldNumber(rhs.mFieldNumber),
			mIndexOptions(rhs.mIndexOptions),
			mFieldName(rhs.mFieldName)
{
	if (!rhs.isStandalone()) {
		dbTable* theTable = dbTable::sCurrentlyConstructing;
		theTable->attachfield(this);
	}
}


dbField::dbField(dbField::FieldIsStandaloneT) :
				mBackend(0),
				mTable(0),
				mFieldNumber(0),
				mIndexOptions(kNotIndexed)
{
} 


dbField::~dbField()
{
	deleteStandaloneStorage();
}


void 
dbField::setName(const char* name)
{ 
	if (isStandalone()) {  // doesn't have name already, almost certainly calling this from
						// inside a dbTable constructor
		deleteStandaloneStorage();
		assert(dbTable::sCurrentlyConstructing);
		dbTable* theTable = dbTable::sCurrentlyConstructing;
		theTable->attachfield(this);
	}
	mFieldName = name; 
}


const char*
dbField::tableName() const
{ 
	if (isStandalone())
		return &OOF_String::sEmptyStr;
	else
		return  mTable->tableName(); 
}


void 
dbField::index(const OOF_IndexOptions indexOpt) 
{
// allowed to turn off existing index
	assert((indexOpt==kNotIndexed)  || (mIndexOptions==kNotIndexed) || (mIndexOptions==indexOpt));		// NOT ALREADY SPECIFIED as different index
	assert(dbConnect::underConstruction());  // prevents call after the time at which it has effect, ie: before newConnection or openConnection
  mIndexOptions=indexOpt;
} 


 
unsigned long 
dbField::fieldEditLen() const
{
		if (fieldType()==charField)
			return fieldLen();
		else
			if (fieldType()==textField)
				return 20;
			else
				return 10;
}


const char* 
dbField::asChars() const
{
	return &OOF_String::sEmptyStr;
}


const void* 
dbField::binaryContents() const
{
	assert(!isStandalone());
	{
		mTable->validateContext();
		return mBackend->getFieldReadFrom(mFieldNumber);
	}
}


OOF_fieldTypes 
dbField::nativeType() const
{
	return fieldType();  // only override for things like dbDate which can be treated as a different native type (ulong)
}


dbTable* 
dbField::fieldTable() const
{
	return mTable;  
}


bool 
dbField::fieldIsBlob() const
{
	return false;  
}


bool 
dbField::fieldIsBinary() const
{
	return true;  
}


bool 
dbField::fieldIsVirtual() const
{
	return false;
}
	
	
bool 
dbField::fieldIsIndexed() const
{
	return mIndexOptions!=kNotIndexed;
} 


bool 
dbField::fieldIsUniqueIndexed() const
{
	return mIndexOptions==kIndexNoDups;
} 
 

char *
dbField::copyAsChars() const
{
	ostrstream oss;
	extract(oss);
	oss << ends;
#ifdef _RTLDLL  // Borland use this to indicate runtime lib in DLL which may mean different allocator
	char* s = oss.str();  // can't just return this - causes crash on delete[]
	long sLen = strlen(s);
	char* ret = new char[sLen+1];
	assert(ret);
	strncpy(ret, s, sLen);
   ret[sLen]='\0';
	oss.rdbuf()->freeze(0);  // leave buffer for stream to delete
	return ret;
#else
	return oss.str();
#endif
}


void 
dbField::extract(ostream& os) const
{
	os << "";
}


bool 
dbField::insert(istream& is, char fieldSep, char recSep)
{
	char c;
	while (is.good()) {
		is >> c;  // just swallowing this character
		if ((c==fieldSep) || (c==recSep)) {
			is.putback(c); // put the separator back so the caller knows what kind we hit
			return true;  // OK - hit a separator
		}
	}
	return false;  // fell through without hitting separator
}


istream& 
operator>>(istream& is, dbField* fld)
{
	fld->insert(is, '\t', '\n');
	return is;
}


istream& 
operator>>(istream& is, dbField& fld)
{
	fld.insert(is, '\t', '\n');
	return is;
}


void 
dbField::describe(ostream& os)
{
	if (mFieldName.isEmpty())
		os << "Un-named field";
	else
		os << "field: " << mFieldName;
	os << " is field type: " << fieldType() << endl;
	if (fieldIsIndexed())
		os << "    has indexing: " << fieldIndexOptions() << endl;
}


// -------------------------------------------------------
//                  d b F i x e d B i n a r y
// -------------------------------------------------------
dbFixedBinary::dbFixedBinary(long fieldWidth) :
	mBodyLength(fieldWidth) 
{}


dbFixedBinary::dbFixedBinary(long fieldWidth, const OOF_IndexOptions indexOptions) :
	dbField(indexOptions),
	mBodyLength(fieldWidth)
{}


dbFixedBinary::dbFixedBinary(long fieldWidth, const char *fieldName)  :
	dbField(fieldName),
	mBodyLength(fieldWidth)
{}


dbFixedBinary::dbFixedBinary(long fieldWidth, const char *fieldName, 
								const OOF_IndexOptions indexOptions)  :
	dbField(fieldName, indexOptions),
	mBodyLength(fieldWidth)
{}


OOF_fieldTypes 
dbFixedBinary::fieldType() const
{
	return fixedBinaryField; 
}


unsigned long 
dbFixedBinary::fieldLen() const
{
	return mBodyLength; 
}


dbFixedBinary& 
dbFixedBinary::operator=(const dbFixedBinary& rhs)
{
	if (this == &rhs)  // MUST use this form to avoid invoking the operator== used for searches
		return *this;
	
	setValue(rhs.binaryContents());
	return *this;
}


dbFixedBinary&  
dbFixedBinary::operator=(const char* str)
{
	assert(strlen(str) <= mBodyLength);
	setValue(str);
	return *this;
}



void 
dbFixedBinary::setValue(const void* str)
{
	assert(!isStandalone());
	mTable->validateContext();
	char *writeTo = (char*)mBackend->getFieldWriteDest(mFieldNumber);
	if (writeTo) {
		memcpy(writeTo, str, mBodyLength);
#ifdef OOF_SmartHeap
	assert(MemPoolCheck(MemDefaultPool));
#endif
	}	
	else
		;  // error NOT YET IMPLEMENTED
	mBackend->markDirty();
}


void 
dbFixedBinary::setByte(char c, unsigned long offset)
{
	assert(!isStandalone());
	mTable->validateContext();
	char *writeTo = (char*) mBackend->getFieldWriteDest(mFieldNumber);
	if (writeTo) {
		writeTo[offset] = c;
#ifdef OOF_SmartHeap
	assert(MemPoolCheck(MemDefaultPool));
#endif
	}	
	else
		;  // error NOT YET IMPLEMENTED
	mBackend->markDirty();
}


void 
dbFixedBinary::extract(ostream& os) const
{
	const char *asChars = (char*)binaryContents();
	os << hex;
	for (unsigned int i=0; i<mBodyLength; i++)
		os << asChars[i] << ' ';
}



void 
dbFixedBinary::generateTestData(const char */*testBuf*/, unsigned long /*testBufLen*/)
{
	for (unsigned int i=0; i<mBodyLength; i++) {
		char c = rand()%0xFF;
		setByte(c, i);
	}
}


void 
dbFixedBinary::describe(ostream& os)
{
	dbField::describe(os);
	os << "    and a length of : " << mBodyLength << endl;
}


// -------------------------------------------------------
//                      d b C h a r
// -------------------------------------------------------
dbChar::dbChar(long fieldWidth) :
				mMaxLength(fieldWidth) 
{
	CompleteInit();
}


dbChar::dbChar(long fieldWidth, const char *fieldName):
				dbField(fieldName),
				mMaxLength(fieldWidth)
{
	CompleteInit();
}


dbChar::dbChar(long fieldWidth, const OOF_IndexOptions indexOptions) :
				dbField(indexOptions),
				mMaxLength(fieldWidth)
{
	CompleteInit();
}



dbChar::dbChar(long fieldWidth, const char *fieldName, const OOF_IndexOptions indexOptions)  :
				dbField(fieldName, indexOptions),
				mMaxLength(fieldWidth)
{
	CompleteInit();
}


dbChar::dbChar(const dbChar& rhs) : 
				dbField(rhs), 
				mMaxLength(rhs.mMaxLength) 
{
	CompleteInit();
	if (isStandalone()) {
		strcpy((char*) mBackend, (const char*) rhs.mBackend);
	}
}
 

dbChar::dbChar(dbField::FieldIsStandaloneT k, long fieldWidth) :
				dbField(k),
				mMaxLength(fieldWidth) 
{
	CompleteInit();
}

 
void
dbChar::CompleteInit()
{
	if (isStandalone())
		mBackend = (OOF_tableBackend*) (new char[mMaxLength+1]);
}
 
 
dbChar::~dbChar()
{
}


void
dbChar::deleteStandaloneStorage()
{
	 if (isStandalone())
		delete[] (char*) mBackend;
}


char * 
dbChar::copyAsChars() const
{
	const char* from = (const char*) *this;
	char* ret = new char[strlen(from)+1];
	assert(ret);
	strcpy(ret, from);
	return ret;
}


void 
dbChar::copyValueFrom(const dbField* srcField)
{
	if (srcField->fieldType()==charField)
		*this = *(const dbChar*)srcField;  // safe downcast
#ifdef OOF_Debug
	else
		dbConnect::raise(ostrstream() << "wrong field type passed in to dbChar::copyValueFrom, copying to " << fieldName() << " from " << srcField->fieldName());
#endif
}


OOF_fieldTypes 
dbChar::fieldType() const
{
	return charField; 
}


bool 
dbChar::fieldIsBinary() const
{
	return false;  
}


unsigned long 
dbChar::fieldLen() const
{
	return mMaxLength; 
}


unsigned short 
dbChar::length() const
{
	if (isStandalone()) {
		return strlen((const char*) mBackend);
	}
	else {
		mTable->validateContext();
		return strlen(*this);
	}
}


const char*
dbChar::asChars() const
{
	return (const char*) *this;
}


dbChar::operator const char*() const
{
	if (isStandalone()) {
		return (const char*) mBackend;
	}
	else {
		mTable->validateContext();
		return (const char*) (mBackend->getFieldReadFrom(mFieldNumber) );
	}
}


#ifdef _Macintosh
void 
dbChar::asStr255(Str255 s) const
{
	const char* from = (const char*) *this;
	s[0] = strlen(from);
	memcpy(s+1, from, s[0]);
#ifdef OOF_SmartHeap
	assert(MemPoolCheck(MemDefaultPool));
#endif
}

void 
dbChar::setStr255(const Str255 s)
{
	long charLen = s[0];
	const char* strStart = (char*) s+1;
	setChars(strStart, charLen);
}



// end of Mac-specific extra conversions
#endif


dbChar& 
dbChar::operator=(const dbChar& rhs)
{
	if (this == &rhs)  // MUST use this form to avoid invoking the operator== used for searches
		return *this;
	
	*this = (const char*)rhs;
	return *this;
}


dbChar&  
dbChar::operator=(const char* str)
{
	setChars(str, strlen(str));
	return *this;
}



void 
dbChar::setChars(const char* str , unsigned long charLen)
{
	if (charLen > mMaxLength)
		;  // error NOT YET IMPLEMENTED
	else
	{
		char *writeTo = 0;
		if (isStandalone()) {
			delete[] (char*) mBackend;
			writeTo = new char[charLen+1];
			mBackend = (OOF_tableBackend*) writeTo;
		}
		else {
			mTable->validateContext();
			writeTo = (char*)mBackend->getFieldWriteDest(mFieldNumber);
			mBackend->markDirty();
		}
		if (writeTo) {
			memcpy(writeTo, str, charLen);
	#ifdef OOF_SmartHeap
		assert(MemPoolCheck(MemDefaultPool));
	#endif
			writeTo[charLen]='\0';
		}
		else
			;  // error NOT YET IMPLEMENTED
	}
}



char& 
dbChar::operator[](unsigned short charIndex)
{
	char *writeTo = 0;
	if (isStandalone()) {
		writeTo =  (char*) mBackend;
	}
	else {
		mTable->validateContext();
		writeTo = (char*)mBackend->getFieldWriteDest(mFieldNumber);
	}
	return writeTo[charIndex];
}


void 
dbChar::extract(ostream& os) const
{
	os << (const char*)*this;
}



bool 
dbChar::insert(istream& is, char fieldSep, char recSep)
{
// this is NOT very efficient, but it *is* safe
	if (!is.good()) 
		return false;
		
	const unsigned long maxLen = fieldLen();
	assert(maxLen);  // brief flash of paranoia in case they get this far with a field of 0 len
	char* buildBuf = new char[maxLen+1];
	assert(buildBuf);
	unsigned int charsRead = 0;
	bool hitEndOfField = false;
	while (is.good() && (charsRead<=maxLen)) {  // allow read one char past, in case it's terminator
		char c;
		is >> c;
		if ((c==fieldSep) || (c==recSep)) {
			is.putback(c); // put the separator back so the caller knows what kind we hit
			hitEndOfField = true;
			break;  // OK - hit a separator
		}
		if ( (c>=' ') && (charsRead<maxLen)) { // printable & in field length
			buildBuf[charsRead] = c;
			charsRead++;
		}
	} // loop reading char at a time

	bool gotDataOK = false;
	if (hitEndOfField)  // everybody happy
		gotDataOK = true;
	else {  // some form of problems along the way
		if (charsRead>maxLen) {// too wide for DB field, store as short anyway
			gotDataOK = true;
			dbField::insert(is, fieldSep, recSep);  // consume the remainder
#ifdef OOF_Debug
			dbConnect::raise(ostrstream() << "dbChar::insert() of field " 
																		<< tableName() << "." << fieldName()
			                              << " has filled its length of " << maxLen 
																		<< " before reaching the end of the field"
											);   
#endif
		}
	}
	
	if (gotDataOK) { 
		buildBuf[charsRead] = '\0';  // now make it into a C string!
		*this = buildBuf;  // finally, assign to the record
	}
	delete[] buildBuf;
	return gotDataOK;  // if general stream problem, let caller report it
}


void 
dbChar::generateTestData(const char *testBuf, unsigned long testBufLen)
{
	unsigned long randLen = rand()%mMaxLength;
	const char *readFrom = testBuf + rand()%(testBufLen-randLen);
	setChars(readFrom, randLen);
}


dbQueryBinary	
dbChar::operator==(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::equals, new dbQueryLiteralStr(str));
}  


dbQueryBinary	
dbChar::startsWith(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::startsWith, new dbQueryLiteralStr(str));
}  


/*

// DOESN'T WORK JUST CREATING STRING IN HERE WITH WILDCARDS
// WILL NEED MORE COMPLEX OPTION

dbQueryBinary	
dbChar::contains(const char* str) const
{
	bool ret;
	char* strPlus;
	unsigned short len = strlen(str);
	if (len==0) {
		strPlus = new char[2];
		strPlus[0] = '*';
		strPlus[1] = '\0';
	}
	else {
		strPlus = new char[len+3];
		assert(strPlus);
		strPlus[0] = '*';
		memcpy(&strPlus[1], str, len);
		strPlus[len+1] = '*';
		strPlus[len+2] = '\0';
	}
	ret = dbQueryBinary(new dbQueryField(this), dbQueryClause::equals, new dbQueryLiteralStr(strPlus));
	delete[] strPlus;
	return ret;

} 
*/ 


dbQueryTrinary	
dbChar::between(const char* fromStr, const char* toStr) const
{
	return dbQueryTrinary(new dbQueryField(this), dbQueryClause::between,
		new dbQueryLiteralStr(fromStr), new dbQueryLiteralStr(toStr));
}


dbQueryTrinary	
dbChar::outside(const char* fromStr, const char* toStr) const
{
	return dbQueryTrinary(new dbQueryField(this), dbQueryClause::outside,
		new dbQueryLiteralStr(fromStr), new dbQueryLiteralStr(toStr));
}


dbQueryBinary	
dbChar::operator!=(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::notEquals, new dbQueryLiteralStr(str));
}  


 dbQueryBinary	
 dbChar::operator<(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::lessThan, new dbQueryLiteralStr(str));
}  


 dbQueryBinary	
 dbChar::operator<=(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::lessThanOrEqual, new dbQueryLiteralStr(str));
}  


 dbQueryBinary	
 dbChar::operator>(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::greaterThan, new dbQueryLiteralStr(str));
}  


 dbQueryBinary	
 dbChar::operator>=(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::greaterThanOrEqual, new dbQueryLiteralStr(str));
}  


void 
dbChar::describe(ostream& os)
{
	dbField::describe(os);
	os << "    and a maximum length of : " << mMaxLength << endl;
}


// -------------------------------------------------------
//                      d b T e x t
// -------------------------------------------------------
dbText::dbText(const OOF_IndexOptions opt) :
	dbBLOB(opt) 
{
//				assert( opt == kNotIndexed);  // more later
}


dbText::dbText(const char *fieldName)  :
	dbBLOB(fieldName) 
{}

 
dbText::dbText(const char *fieldName, const OOF_IndexOptions opt)  :
	dbBLOB(fieldName, opt) 
{
//assert(opt == kNotIndexed);
}


dbText::operator const char*() const
{
	if (isStandalone()) {
		if (fieldLen()>0)
			return (char *)bodyAddress();
		else
			return &OOF_String::sEmptyStr;
	}
	else {
		if (!loaded())
			LoadField();
		if (fieldDataLen())
			return (char *)bodyAddress();
		else
			return &OOF_String::sEmptyStr;
	}
}


const char*
dbText::asChars() const
{
	return (const char*) *this;
}


char * 
dbText::copyAsChars() const
{
	const char* from = (const char*) *this;
	char* ret;
	unsigned long len = strlen(from);
	if (len) {
		ret = new char[len+1];  
		memcpy(ret, from, len);
#ifdef OOF_SmartHeap
	assert(MemPoolCheck(MemDefaultPool));
#endif
	}
	else {
		ret = new char[1];
		ret[0] = '\0';
	}
	return ret;
}


void 
dbText::copyValueFrom(const dbField* srcField)
{
	if (srcField->fieldType()==textField)
		*this = *(dbText*)srcField;  // safe downcast
#ifdef OOF_Debug
	else
		dbConnect::raise(ostrstream() << "wrong field type passed in to dbText::copyValueFrom, copying to " << fieldName() << " from " << srcField->fieldName());
#endif
}


OOF_fieldTypes 
dbText::fieldType() const
{
	return textField; 
}


OOF_fieldTypes 
dbText::nativeType() const
{
	return charField; 
}


bool 
dbText::fieldIsBinary() const
{
	return false;  
}


dbText& 
dbText::operator=(const dbText& rhs)
{
	if (this == &rhs)  // MUST use this form to avoid invoking the operator== used for searches
		return *this;
	
	*this = (const char*)rhs;
	return *this;
}


dbText& 
dbText::operator=(const char* str)
{
	unsigned long charLen;
	if (str)
		charLen = strlen(str);
	else
		charLen = 0;
	if (charLen) {
		allocRoomFor(charLen+1);
		memcpy((char*)bodyAddress(), str, charLen);
		((char*)bodyAddress())[charLen] = '\0';
#ifdef OOF_SmartHeap
	assert(MemPoolCheck(MemDefaultPool));
#endif
		SetLength(charLen+1);
	}
	else
		SetLength(0);
	mBackend->markDirty();
	markDirty();
	return *this;
}


void 
dbText::operator+=(const char* str)
{
	unsigned long charLen = strlen(str);
	if (charLen==0)
		return;
		
	if (isStandalone()) {
		const char* from = (const char*) *this;
		unsigned long newLen = strlen(from)+charLen;
		unsigned long currentFldLen = 0;
		if (bodyAddress())
			currentFldLen = strlen((char*)bodyAddress());
		allocRoomFor(newLen);
		memcpy((char*)bodyAddress()+currentFldLen-1, str, charLen);	// we KNOW where the null is
		((char*)bodyAddress())[newLen-1] = '\0';
		#ifdef OOF_SmartHeap
			assert(MemPoolCheck(MemDefaultPool));
		#endif		
	}
	else {
		unsigned long currentFldLen = mBackend->fieldDataLen(this);	// inc null
		unsigned long newLen = currentFldLen+charLen;
		allocRoomFor(newLen);
		if (!loaded())
			LoadField();
		memcpy((char*)bodyAddress()+currentFldLen-1, str, charLen);	// we KNOW where the null is
		((char*)bodyAddress())[newLen-1] = '\0';
		#ifdef OOF_SmartHeap
			assert(MemPoolCheck(MemDefaultPool));
		#endif
		SetLength(newLen);
		mBackend->markDirty();
		markDirty();		
	}
}


void 
dbText::extract(ostream& os) const
{
	os << (const char*)*this;
}


bool 
dbText::insert(istream& is, char fieldSep, char recSep)
{
// nearly straight copy of dbChar - we really need to do something to combine these fields!

// this is NOT very efficient, but it *is* safe
	if (!is.good()) 
		return false;

// this is not yet complete - it needs a cleaner way to handle large imports		
	const unsigned long maxLen = 16000;  // KEEP UNDER 64k to avoid Win3.1 bugs
	char* buildBuf = new char[maxLen+1];
	assert(buildBuf);
	unsigned int charsRead = 0;
	bool hitEndOfField = false;
	while (is.good() && (charsRead<=maxLen)) {  // allow read one char past, in case it's terminator
		char c;
		is >> c;
		if ((c==fieldSep) || (c==recSep)) {
			is.putback(c); // put the separator back so the caller knows what kind we hit
			hitEndOfField = true;
			break;  // OK - hit a separator
		}
		if ( (c>=' ') && (charsRead<maxLen)) { // printable & in field length
			buildBuf[charsRead] = c;
			charsRead++;
		}
	} // loop reading char at a time

	bool gotDataOK = false;
	if (hitEndOfField)  // everybody happy
		gotDataOK = true;
	else {  // some form of problems along the way
		if (charsRead>maxLen) {// too wide for DB field, store as short anyway
			gotDataOK = true;
			dbField::insert(is, fieldSep, recSep);  // consume the remainder
#ifdef OOF_Debug
			dbConnect::raise(ostrstream() << "dbText::insert() of field " 
																		<< tableName() << "." << fieldName()
			                              << " has filled its length of " << maxLen 
																		<< " before reaching the end of the field"
											);   
#endif
		}
	}
	
	if (gotDataOK) { 
		buildBuf[charsRead] = '\0';  // now make it into a C string!
		*this = buildBuf;  // finally, assign to the record
	}
	else
		*this = "";
	delete[] buildBuf;
	return gotDataOK;  // if general stream problem, let caller report it
}


void 
dbText::generateTestData(const char *testBuf , unsigned long testBufLen)
{
	unsigned long randLen = rand()%testBufLen%10240;		// up to 10kb allowed in one dbText
	const char *readFrom = testBuf + rand()%(testBufLen-randLen);
	setChars(readFrom, randLen);
}


dbQueryBinary	
dbText::operator==(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::equals, new dbQueryLiteralStr(str));
}  

	
dbQueryBinary	
dbText::operator!=(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::notEquals, new dbQueryLiteralStr(str));
}  


dbQueryBinary	
dbText::operator<(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::lessThan, new dbQueryLiteralStr(str));
}  


dbQueryBinary	
dbText::operator<=(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::lessThanOrEqual, new dbQueryLiteralStr(str));
}  


dbQueryBinary	
dbText::operator>(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::greaterThan, new dbQueryLiteralStr(str));
}  


dbQueryBinary	
dbText::operator>=(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::greaterThanOrEqual, new dbQueryLiteralStr(str));
}  



// -------------------------------------------------------
//                      d b B L O B s t o r a g e
// -------------------------------------------------------
dbBLOBstorage::dbBLOBstorage(dbBLOB* owner)  :
				mBody(0),
				mBodyLength(0),
				mDirty(false),
				mLoaded(false),
				mOwner(owner)
	{};  
	
dbBLOBstorage::~dbBLOBstorage()
{
	delete[] mBody;
};


void *
dbBLOBstorage::bodyAddress() const
{
	return mBody;
}


bool 
dbBLOBstorage::isDirty() const
{
	return mDirty;
}

void
dbBLOBstorage::markDirty()
{
	mDirty = true;
	mLoaded = true;
}


void 
dbBLOBstorage::reset()
{
	mDirty = false;
	mLoaded = false;
}


void* 
dbBLOBstorage::orphanBody()
{
	
	void* ret = mBody;
	reset();
	mBody = 0;
	mBodyLength = 0;
	return ret;
}


void 
dbBLOBstorage::adoptBody(void* body, unsigned long bodyLen)
{
	delete[] mBody;
	mBody = body;
	mBodyLength = bodyLen;
	mLoaded = mDirty = true;
}


void* 
dbBLOBstorage::allocRoomFor(unsigned long len)
{
	if (len > mBodyLength) {
		if (mLoaded) {	// assume we must preserve the buffer
			char *newBody = new char[len];
			assert(newBody);
			if (mBody) {
				memcpy(newBody, mBody, mBodyLength);
				delete[] mBody;
			}
			mBody = newBody;
		}
		else {			
			delete[] mBody;
			mBody = new char[len];
			assert(mBody);
		}
		mBodyLength = len;
	}
#ifdef OOF_SmartHeap
	assert(MemPoolCheck(MemDefaultPool));
#endif
	return mBody;
}


void 
dbBLOBstorage::setChars(const char* str , unsigned long charLen)
{
	if (str && charLen) {
		allocRoomFor(charLen+1);
		memcpy(mBody, str, charLen);
#ifdef OOF_SmartHeap
	assert(MemPoolCheck(MemDefaultPool));
#endif
		((char*)mBody)[charLen]='\0';
		mOwner->SetLength(charLen+1);
	}
	else
		mOwner->SetLength(0);
	mOwner->markDirty();
}


// -------------------------------------------------------
//                      d b B L O B
// -------------------------------------------------------
dbBLOB::dbBLOB()
{
	mStorage = new dbBLOBstorage(this);
}


dbBLOB::dbBLOB(const OOF_IndexOptions indexOptions) :
				dbField(indexOptions)
{
	mStorage = new dbBLOBstorage(this);
}


dbBLOB::dbBLOB(const char *fieldName)  :
				dbField(fieldName)
{
	mStorage = new dbBLOBstorage(this);
}


dbBLOB::dbBLOB(const char *fieldName, const OOF_IndexOptions indexOptions)  :
				dbField(fieldName, indexOptions)
{
	mStorage = new dbBLOBstorage(this);
}


dbBLOB::dbBLOB(const dbBLOB& rhs) : dbField(rhs)
{
	mStorage = new dbBLOBstorage(this);
}


dbBLOB::~dbBLOB()
{
	delete mStorage;
};


OOF_fieldTypes 
dbBLOB::fieldType() const
{
	return blobField; 
}


unsigned long 
dbBLOB::fieldLen() const
{
	if (isStandalone())
		if (mStorage)
			return mStorage->mBodyLength;
		else
			return 0;
	else
		return mBackend->blobPointerReferenceSize();  
}

 
bool 
dbBLOB::fieldIsBlob() const
{
	return true;  
}


void 
dbBLOB::SetLength(unsigned long len)
{
	assert(!isStandalone());
	mTable->validateContext();
	mBackend->setBlobLength(mFieldNumber, len);
}


void 
dbBLOB::LoadField() const
{
	assert(!isStandalone());
	mTable->validateContext();
	mBackend->loadBlob(this);
}


// -------------------------------------------------------
//               d b C o m p o u n d F i e l d 
// -------------------------------------------------------
dbCompoundField::dbCompoundField() :
	mTotalLength(0)
{}
	
								
dbCompoundField::dbCompoundField(const OOF_IndexOptions indexOptions) :
	dbField(indexOptions),
	mTotalLength(0)
{}
	
								
dbCompoundField::dbCompoundField(const char *fieldName) :
	dbField(fieldName),
	mTotalLength(0)
{}
	
								
dbCompoundField::dbCompoundField(const char *fieldName, const OOF_IndexOptions indexOptions) :
	dbField(fieldName, indexOptions),
	mTotalLength(0)
{}
	
								
dbCompoundField::dbCompoundField(const dbCompoundField& rhs) : 
	dbField(rhs), 
	OOF_FieldSegOwner(rhs),  
	mTotalLength(rhs.mTotalLength)
{}


OOF_fieldTypes 
dbCompoundField::fieldType() const
{
	return compoundField;
}


unsigned long 
dbCompoundField::fieldLen() const
{
	return mTotalLength;   
}


bool 
dbCompoundField::fieldIsVirtual() const
{
	return true;
}


void 
dbCompoundField::addSegment(dbField& segField) 
{
	assert(segField.fieldType() != textField);
	OOF_FieldSegOwner::addSegment(segField);
	mTotalLength += segField.fieldLen();
}


const dbField* 
dbCompoundField::field(unsigned int n) const
{
	const dbField* theField = mTable->field(OOF_FieldSegOwner::fieldNumber(n));
	return theField;
}


dbQueryBinary	
dbCompoundField::operator==(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::equals, new dbQueryLiteralStr(str));
}  


dbQueryBinary	
dbCompoundField::operator!=(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::notEquals, new dbQueryLiteralStr(str));
}  


dbQueryBinary	
dbCompoundField::operator<(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::lessThan, new dbQueryLiteralStr(str));
}  


dbQueryBinary	
dbCompoundField::operator<=(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::lessThanOrEqual, new dbQueryLiteralStr(str));
}  


dbQueryBinary	
dbCompoundField::operator>(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::greaterThan, new dbQueryLiteralStr(str));
}  


dbQueryBinary	
dbCompoundField::operator>=(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::greaterThanOrEqual, new dbQueryLiteralStr(str));
}  


dbQueryBinary	
dbCompoundField::startsWith(const char* str) const
{
	return dbQueryBinary(new dbQueryField(this), dbQueryClause::startsWith, new dbQueryLiteralStr(str));
}  


void 
dbCompoundField::extract(ostream& os) const
{
	unsigned int numSegments = countSegments();
	for (unsigned int i=0; i<numSegments; i++) {
		const dbField* fld = field(i);
		fld->extract(os);
	}
}

 
// -------------------------------------------------------
//                  O O F I L E _ f i e l d T y p e s
// -------------------------------------------------------
ostream& 
operator<<(ostream& os, OOF_fieldTypes ft)
{
	switch (ft) {
	case (charField) :
		os << "fixed-length chars";
		break;
		
	case (textField) :
		os << "variable-length text";
		break;
		
	case (dateField) :
		os << "date";
		break;
		
	case (shortField) :
		os << "short integer";
		break;
		
	case (uShortField) :
		os << "unsigned short integer";
		break;
		
	case (longField) :
		os << "long integer";
		break;
		
	case (uLongField) :
		os << "unsigned long integer";
		break;
		
	case (realField) :
		os << "real";
		break;
		
	case (blobField) :
		os << "binary blob";
		break;
		
	case (compoundField) :
		os << "compound";
		break;
		
	case (fixedBinaryField) :
		os << "compound";
		break;
		
	case (boolField) :
		os << "compound";
		break;
		
	case (relationshipField) :
		os << "relationship traversal pointer";
		break;
		
	}
	return os;
}


// -------------------------------------------------------
//                d b F i e l d C o p i e r
// -------------------------------------------------------
dbFieldCopier::dbFieldCopier(dbBroadcaster* listensTo, unsigned short copyOnMessage, dbField* copyFrom, dbField* copyTo) :
		dbSingleListener(listensTo),
		mCopyFrom(copyFrom),
		mCopyTo(copyTo),
		mCopyOnMessage(copyOnMessage)
{}


void
dbFieldCopier::receiveMsg(OOFmsgT msg, unsigned long senderDefined)
{
	if (msg==mCopyOnMessage)
		mCopyTo->copyValueFrom(mCopyFrom);
	else
		switch (msg) {
		case (OOFmsg_BroadcasterClosing) : 
			mListensTo = 0;  // won't be around to tell it that we're being deleted
			delete this;
			break;
	
		default :
			dbSingleListener::receiveMsg(msg, senderDefined);
		};
}


// -------------------------------------------------------
//               dbCloneCleanup
// -------------------------------------------------------
dbCloneCleanup::dbCloneCleanup(dbTable* listensTo) :
	dbSingleListener(listensTo),
	mTable(listensTo)
{}


void
dbCloneCleanup::receiveMsg(OOFmsgT msg, unsigned long /* senderDefined */)
{
	if (msg!=OOFmsg_BroadcasterClosing)  
		mTable->postCloneCleanup();
	delete this;
}






// -------------------------------------------------------
//               O O F _ S e g m e n t  
// -------------------------------------------------------

OOF_Segment::OOF_Segment(const dbField& fld) : 
		mFieldNo(fld.fieldNumber()), 
		mReversed(false) 
{}


OOF_Segment::OOF_Segment(const dbField* fld) : 
		mFieldNo(fld->fieldNumber()), 
		mReversed(false) 
{}
