#ifndef H_OOFquery
#define H_OOFquery

// COPYRIGHT 1994 A.D. Software, All rights reserved

// IMPLEMENTATION of queries
// known to Fields and Query implementations

#include "oofbool.hpp"

class dbQueryField ;
class dbQueryLiteral;
class dbQueryLiteralStr;
class dbQueryBinaryCombo;
class dbQueryBinaryComboOwner;
class dbField;
class dbCompoundField;

class dbQueryClause {
public:
	enum BinaryQueryOps {equals, notEquals, lessThan, lessThanOrEqual, greaterThan, greaterThanOrEqual, startsWith};
	enum TrinaryQueryOps {between, outside};
	enum QueryCombinatorialOps {and, or};
	enum QueryClauseTypes {binaryfieldToLiteral, binaryfieldTofield, trinaryFieldToLiterals, atomicfield, atomicLiteral, atomicLiteralStr, binaryCombination, builtQuery};

	// use default copy ctor
	virtual ~dbQueryClause() {};
	
	virtual dbQueryClause::QueryClauseTypes  queryClauseType() const=0;
	bool isSimpleIndexedSearch() const;
	
// combinatorial operators
	dbQueryBinaryCombo operator&(const dbQueryClause&) const;
	dbQueryBinaryCombo operator&&(const dbQueryClause&) const;
	dbQueryBinaryCombo operator|(const dbQueryClause&) const;
	dbQueryBinaryCombo operator||(const dbQueryClause&) const;

	dbQueryBinaryComboOwner* operator|(dbQueryClause* adoptRHS);
	dbQueryBinaryComboOwner* operator||(dbQueryClause* adoptRHS);
	dbQueryBinaryComboOwner* operator&(dbQueryClause* adoptRHS);
	dbQueryBinaryComboOwner* operator&&(dbQueryClause* adoptRHS);
	
protected:
	dbQueryClause() {};
	
// other protocol
// part of testing	dbTable *AppliesToTable();

public:
	static bool wildcardsInString(const char*);
	static char sWildcardMultiple;  // not const so user can change!
	static char sWildcardSingle;
};
//operator=(dbTable lhs, dbQueryClause& rhs);


// class for building queries
class dbQuery : public dbQueryClause {
public:
	dbQuery();
	// use default copy ctor
	virtual ~dbQuery();
	
	virtual dbQueryClause::QueryClauseTypes  queryClauseType() const;

	void operator|=(dbQueryClause* adoptedPortion);
	void operator&=(dbQueryClause* adoptedPortion);
	
	const dbQueryClause* query() const;
	
private:
	dbQueryClause* mQuery;  // recursively built as binary combos
};


class dbQueryBinary : public dbQueryClause {
public:
	dbQueryBinary(dbQueryField* fld, BinaryQueryOps op, dbQueryClause* litOrFld);
	dbQueryBinary(dbQueryBinary&);  // CAN'T USE CONST RHS
	virtual ~dbQueryBinary();
private:
	dbQueryBinary& operator=(dbQueryBinary&);

public:
	virtual dbQueryClause::QueryClauseTypes queryClauseType()  const;
	dbQueryLiteralStr* literalStrClause() const;
	dbQueryLiteral* literalClause() const;
	dbQueryClause::BinaryQueryOps binaryOperator() const;
	const dbField* lhsField() const;
	bool isSimpleIndexedSearch() const;
	
// for evaluating comparisons in if()
//	operator bool() const; too hard to evaluate in 1.1, will be finished in 1.2

protected:
	
// other protocol
// part of testing	dbTable *AppliesToTable();

// data storage
protected:
	dbQueryField  *mLhs;  // owned
	dbQueryClause* mRhs;  // owned
	BinaryQueryOps  mBinOp;
};


class dbQueryBinaryCombo : public dbQueryClause {
public:
// NOT YET IMPLEMENTED extend to any number of clauses with same op
	dbQueryBinaryCombo(const dbQueryClause*, QueryCombinatorialOps, const dbQueryClause*);
	dbQueryBinaryCombo(const dbQueryBinaryCombo&);
	virtual ~dbQueryBinaryCombo();
private:
	dbQueryBinaryCombo& operator=(dbQueryBinaryCombo&);

public:
	virtual dbQueryClause::QueryClauseTypes queryClauseType()  const;
	dbQueryClause::QueryCombinatorialOps binaryOperator() const;
	const dbQueryClause* item(unsigned int) const;
	unsigned long pairFieldsIfCouldUseCompoundIndex() const;
	dbQueryClause* makeCompoundSearch(const dbCompoundField*);

protected:
	
// other protocol
// part of testing	dbTable *AppliesToTable();

// data storage
protected:
// NOTE unlike the other Binary and Trinary, we do NOT own our lhs & rhs
	const dbQueryClause  *mLhs, *mRhs;
	QueryCombinatorialOps  mComboOp;
	char *mTempKeyStoreToDispose, *mTempKeyStoreToDispose2;
};


class dbQueryBinaryComboOwner : public dbQueryBinaryCombo {
public:
	dbQueryBinaryComboOwner(dbQueryClause*, QueryCombinatorialOps, dbQueryClause*);
	// use default copy ctor
	virtual ~dbQueryBinaryComboOwner();		
private:
	dbQueryClause  *mOurLhs, *mOurRhs;  // owned
};


class dbQueryTrinary : public dbQueryClause {
public:
	dbQueryTrinary(dbQueryField* fld, TrinaryQueryOps op, dbQueryClause* litOrFldFrom, dbQueryClause* litOrFldTo) :
				mLhs(fld),
				mFrom(litOrFldFrom),
				mTo(litOrFldTo),
				mTrinOp(op)
				{};		
	dbQueryTrinary(dbQueryTrinary&);  // CAN'T USE CONST
	virtual ~dbQueryTrinary();
private:
	dbQueryTrinary& operator=(dbQueryTrinary&);

public:
	virtual dbQueryClause::QueryClauseTypes queryClauseType()  const;
	dbQueryLiteralStr* literalStrFromClause() const;
	dbQueryLiteralStr* literalStrToClause() const;
	dbQueryLiteral* literalFromClause() const;
	dbQueryLiteral* literalToClause() const;
	dbQueryClause::TrinaryQueryOps trinaryOperator() const;
	const dbField* lhsField() const;

protected:
	
// data storage
protected:
	dbQueryField  *mLhs;  // owned
	dbQueryClause *mFrom, *mTo;  // owned
	TrinaryQueryOps  mTrinOp;
};


class dbQueryLiteral : public dbQueryClause {
public:
	virtual dbQueryClause::QueryClauseTypes  queryClauseType()  const;
	virtual const void* binaryContents() const;
	virtual unsigned short literalLen() const;
	virtual ~dbQueryLiteral() {};		
	
protected:
	dbQueryLiteral() {};		
	// use default copy ctor
};


class dbQueryLiteralStr : public dbQueryLiteral {
public:
	dbQueryLiteralStr(const char* str) :
	mStr(str),
	mCheckedForWildcards(neverChecked)
	{};
	// use default copy ctor
	virtual ~dbQueryLiteralStr() {};

	virtual dbQueryClause::QueryClauseTypes  queryClauseType()  const;
	const char* str() const;
	virtual const void* binaryContents() const;
	virtual unsigned short literalLen() const;
	
	bool hasWildcards();

// data storage
private:
	const char* mStr;
	enum WildcardCheckT { neverChecked, hasNone, hasWild };
	WildcardCheckT mCheckedForWildcards;
}; 


class dbQueryLiteralDouble : public dbQueryLiteral {
public:
	dbQueryLiteralDouble(double d) :
	mNum(d) {};
	// use default copy ctor
	virtual ~dbQueryLiteralDouble() {};
	
	virtual const void* binaryContents() const;
	virtual unsigned short literalLen() const;

// data storage
private:
	const double mNum;
}; 


class dbQueryLiteralBLOB : public dbQueryLiteral {
// mainly used to support compound key queries ending in binary number
public:
	dbQueryLiteralBLOB(void* blob, unsigned short len) :
										mBLOB(blob),
										mBLOBlen(len)
										{};
	// use default copy ctor
	virtual ~dbQueryLiteralBLOB() {};
	
	virtual const void* binaryContents() const;
	virtual unsigned short literalLen() const;

// data storage
private:
	const void* mBLOB;
	unsigned short mBLOBlen;
}; 


class dbQueryLiteralLong : public dbQueryLiteral {
public:
	dbQueryLiteralLong(long l) :
	mNum(l) {};
	// use default copy ctor
	virtual ~dbQueryLiteralLong() {};
	
	virtual const void* binaryContents() const;
	virtual unsigned short literalLen() const;

// data storage
private:
	const long mNum;
}; 


class dbQueryLiteralUlong : public dbQueryLiteral {
public:
	dbQueryLiteralUlong(unsigned long l) :
	mNum(l) {};
	// use default copy ctor
	virtual ~dbQueryLiteralUlong() {};
	
	virtual const void* binaryContents() const;
	virtual unsigned short literalLen() const;

// data storage
private:
	const unsigned long mNum;
};


class dbQueryLiteralShort : public dbQueryLiteral {
public:
	dbQueryLiteralShort(short s) :
	mNum(s) {};
	// use default copy ctor
	virtual ~dbQueryLiteralShort() {};
	
	virtual const void* binaryContents() const;
	virtual unsigned short literalLen() const;

// data storage
private:
	const short mNum;
};


class dbQueryLiteralUshort : public dbQueryLiteral {
public:
	dbQueryLiteralUshort(unsigned short s) :
	mNum(s) {};
	// use default copy ctor
	virtual~dbQueryLiteralUshort() {};
	
	virtual const void* binaryContents() const;
	virtual unsigned short literalLen() const;

// data storage
private:
	const unsigned short mNum;
};


class dbQueryField : public dbQueryClause {
public:
// constructors
	dbQueryField(const dbField* fld) :
				mfield(fld) {};
	// use default copy ctor
	virtual~dbQueryField() {};
	
	const dbField*  field()  const {return mfield;};
	virtual dbQueryClause::QueryClauseTypes  queryClauseType()  const;
	
// data storage
	const dbField* mfield;
};


// include inline definitions
#include "oofquery.inl"
#endif
