// COPYRIGHT 1994 A.D. Software, All rights reserved

// public layer of OOFILE database
// implementing relations

#include "oof3.hpp"
#include "oofrel.hpp"



// -------------------------------------------------------
//              d b R e l R e f B a s e 
// -------------------------------------------------------
dbRelRefBase::dbRelRefBase(unsigned long maxLinks) : 
				mMinLinks(0),
				mMaxLinks(maxLinks),
				mRelatedTableProto(0),
				mActualRelatedTable(0),
				mInverseFieldNumber(0),
				mJoinField(0),
				mPropagatesDeletes(true),
				mInverseField(0)
{
}


dbRelRefBase::dbRelRefBase(const dbRelRefBase& rhs) : 
				dbField(rhs),
				mMinLinks(rhs.mMinLinks),
				mMaxLinks(rhs.mMaxLinks),
				mRelatedTableProto(rhs.mRelatedTableProto),
				mActualRelatedTable(0),
				mInverseFieldNumber(rhs.mInverseFieldNumber),
				mJoinField(rhs.mJoinField),  // see postCloneCleanup
				mPropagatesDeletes(rhs.mPropagatesDeletes),
				mInverseField(0)
{
}


void
dbRelRefBase::postCloneCleanup() 
{
	// by this stage we are guaranteed that all fields in the table exist and 
	// have been registered with the table, so can pull out our real field from
	// the dictionary.
	if (mJoinField) {
		mJoinField = mTable->field(mJoinField->fieldNumber());
	}
}


void 
dbRelRefBase::SetCommonRelationshipFields(dbRelRefBase& rhs)
{
	mInverseFieldNumber = rhs.fieldNumber();
	mRelatedTableProto = rhs.fieldTable();
	if (rhs.mMaxLinks!=1 && !mJoinField)
		index(kIndexed);  // unless a join field later specified
}


void 
dbRelRefBase::relatesTo(dbRelRefBase& rhs)
{
	SetCommonRelationshipFields(rhs);
	if (!rhs.CompletelySpecified()) {
		rhs.propagateRelatedDeletes(false);  // we own it, not vice versa
		rhs.relatesTo(*this);
	}
}


void 
dbRelRefBase::joinField(dbField& joinBy)
{
	assert(joinBy.fieldTable() == mTable);
	mJoinField = &joinBy;
	index(kNotIndexed);  // only indexed if an N-ary side of an oid relationship
}


void 
dbRelRefBase::BuildRelatedTable()
{
	#ifdef OOF_Debug
		if (!mRelatedTableProto)
			dbConnect::raise(ostrstream() << "Relationship field No " << fieldNumber() 
			<< " in table " << tableName() << " lacks a rhs");
	#else
		assert(mRelatedTableProto);
	#endif
	assert(mActualRelatedTable==0);  // should only be called in this circumstance
	mActualRelatedTable = mRelatedTableProto->cloneWithoutSelection();
	assert(mActualRelatedTable);
//	mActualRelatedTable->setSaveOption(dbTable::autoOnContextChange);
	mActualRelatedTable->setSaveOption(dbTable::requireExplicitAndBuffer);
	
	mInverseField = OurInverse();
	assert(mInverseField);
	
	assert( (mJoinField==0) == (mInverseField->joinField()==0) );  // either both zero or both pointers
	mInverseField->completeCloneOfInverse(this);
}


void
dbRelRefBase::completeCloneOfInverse(dbRelRefBase* cloner) 
{
	mTable->mRelated = this;  
// a table only needs to know mRelated when the table is a clone managing a related selection
	mActualRelatedTable = cloner->fieldTable();
	subscribeTo(mActualRelatedTable); 
// depend on the lhs of the rel for things like context changes, selection changes

	mInverseField = cloner;
}


void 
dbRelRefBase::relateRecord()
{
	assert(mInverseField);
	mInverseField->relateFromRecord();
}


void 
dbRelRefBase::relateSelection()
{
	for (mTable->start(); mTable->more(); mTable->next())
		relateRecord();
}


void 
dbRelRefBase::relateFromRecord()
{
	if (mJoinField)  // is a join relationship
	{
		mBackend->loadRelatedContextJoiningFromTo(mInverseField->joinField(), mJoinField);
	}
	else {
		oidT oidToFind = mInverseField->OID();
		if (mInverseField->mMaxLinks==1)
			mBackend->selectJustOID(oidToFind);
		else {
			assert(fieldIsIndexed());
			mBackend->searchEqual(this, &oidToFind);
		}
	}
}


OOF_fieldTypes 
dbRelRefBase::fieldType() const
{
	return relationshipField; 
}


OOF_fieldTypes 
dbRelRefBase::nativeType() const
{
	if (mJoinField)
		return fieldType();
	else
		return uLongField;
}


bool 
dbRelRefBase::fieldIsVirtual() const
{
	if (mJoinField)
		return true;
	else
		return false;  // stores an OID
}


unsigned long dbRelRefBase::fieldLen() const
{
	if (mJoinField)
		return 0;  // no storage in the record
	else
		return sizeof(oidT); // stores an OID
}


oidT
dbRelRefBase::OID() const
{
	assert(!mJoinField);  // shouldn't be called this is a join relationship
	oidT *readFrom = (oidT *)mBackend->getFieldReadFrom(mFieldNumber);
	oidT currentOID = *readFrom;
// WRONG - use 0 oid to indicate a relationship that's never been traversed
//	if (!currentOID)
//		currentOID = mTable->currentOID();  // if we never stored an OID, use the table
//	assert(currentOID);  // should never be called when this can't be obtained
	return currentOID;	
}


void 
dbRelRefBase::OID(oidT oid)
{
	oidT currentOID = OID();
	if (currentOID) {
		assert(currentOID==oid);  // NOT YET IMPLEMENTED - change relationships
	}
	else {  // we need setting!
		oidT *writeTo = (oidT *)mBackend->getFieldWriteDest(mFieldNumber);
		if (writeTo)
			*writeTo = oid;	
		mBackend->markDirty();
	}
}


void 
dbRelRefBase::receiveMsg(OOFmsgT msg, unsigned long senderDefined)
{
	switch (msg) {
	case (OOFmsg_ChangeSelection) :
		mTable->selectNone();  // don't just unload the record - there are NO related recs!
		break;
		
	case (OOFmsg_ChangeContext) :
		mTable->ContextChange();  // may be propagating caching changes
		break;
		
	case (OOFmsg_UnloadRecord) :
		mTable->selectNone();
		break;
		
	case (OOFmsg_DeleteRecord) :
		if (mInverseField->propagatesDeletes())  // it propagates to us
			mTable->deleteSelection();
		break;
		
	case (OOFmsg_SaveRecord) :
		if (mTable->isRecordLoaded()) {
			mTable->saveRecord();
			if (!mJoinField) {
				if (mInverseField->mMaxLinks==1)  // lhs of 1:something
	// tell our inverse to point straight at us, now we have an oid
					mInverseField->OID(mTable->currentOID());
				else  // lhs of N:something
					if (mMaxLinks==1)  // lhs of N:1
					// other side's OID is itself, used to search us
						mInverseField->OID(mInverseField->fieldTable()->currentOID());
					else  // N:N - need independent OID
						dbConnect::raise("N:N relationships not yet supported");  // later get a relationship OID
			}
		}
		break;
		
	case (OOFmsg_BroadcasterClosing) : 
		mListensTo = 0;  // won't be around to tell it that we're being deleted
		delete mTable;  // yes, delete the table! it's a clone on the heap, managing this relationship
		break;
		
	default :
		dbSingleListener::receiveMsg(msg, senderDefined);
	};
}


void 
dbRelRefBase::updateRelField()
{
// NOTE: updating the lhs relationship is performed in receiveMsg, AFTER our saveRecord
// in case our table's OID is required by the rhs
	if (mJoinField) 
		mJoinField->copyValueFrom(mInverseField->joinField());
	else { // grab an OID
		if (mMaxLinks==1)  // rhs of something:1 rel
			OID(mInverseField->fieldTable()->currentOID());  // we point straight at our inverse
		else  // rhs of something:N rel
			if (mInverseField->mMaxLinks==1)  // rhs of 1:N rel
				OID(mTable->currentOID());  // store our own table's OID to search inverse
			else  // N:N - need independent OID
				dbConnect::raise("N:N relationships not yet supported");  // later get a relationship OID
	}
}


void 
dbRelRefBase::describe(ostream& os)
{
	os << "field: " << fieldName() << " is a ";
	dbRelRefBase* inverseProto = (dbRelRefBase* )mRelatedTableProto->field(mInverseFieldNumber);    // safe downcast
// always exists regardless of relationship status
	
	if (inverseProto->mMaxLinks==1)
		os << "1";
	else
		os << "N";
	
	if (mMaxLinks==1)
		os << ":1";
	else
		os << ":N";
	
	os << " relationship ";
	
	if (mJoinField)
		os << "joining across field: " << mJoinField->fieldName();

	if (propagatesDeletes())
		os << " and propagating deletes to related records";
	os << endl;
}



void 
dbRelRefBase::extract(ostream& os) const
{
	if (mJoinField)
		mJoinField->extract(os);
	else
		os << OID();
}


const OOF_String& 
dbRelRefBase::fieldName() const
{
// override as the table name could change at a later date
// than our initialisation - it's very messy getting the ultimate name
// and this is NOT a function that needs to be particularly efficient
	const OOF_String& tempName = dbField::fieldName();
	if (tempName.isEmpty())
		return mRelatedTableProto->tableName();
	else
		return tempName;
}


// -------------------------------------------------------
//              d b R e l R e f B a s e 
// -------------------------------------------------------
dbRelationship::dbRelationship(dbRelRefBase& lhs, dbRelRefBase& rhs)
{
	lhs.relatesTo(rhs);
}


dbRelationship::dbRelationship(dbRelRefBase* lhs, dbRelRefBase* rhs)
{
	lhs->relatesTo(*rhs);
}


dbRelationship::dbRelationship(dbRelRefBase& lhs, dbRelRefBase& rhs, dbField& lhsJoin, dbField& rhsJoin)
{
	lhs.relatesTo(rhs);
	lhs.joinField(lhsJoin);
	rhs.joinField(rhsJoin);
}


dbRelationship::dbRelationship(dbRelRefBase* lhs, dbRelRefBase* rhs, dbField* lhsJoin, dbField* rhsJoin)
{
	lhs->relatesTo(*rhs);
	lhs->joinField(*lhsJoin);
	rhs->joinField(*rhsJoin);
}
